#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) 2007-2020, Acoular Development Team.
#------------------------------------------------------------------------------
"""

Loads the example data set, sets diffrent Sectors for intergration.
Shows acoular Sector und Sound Pressure level Integration functionallity.

"""

from os import path
from numpy import array,arange
import acoular
from pylab import figure, plot, imshow, colorbar, show,xlim, ylim,legend,cm
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon,Rectangle

#fft block size
block = 128

#example data 
micgeofile = path.join( path.split(acoular.__file__)[0],'xml','array_56.xml')
mg = acoular.MicGeom( from_file=micgeofile )
ts = acoular.TimeSamples( name='example_data.h5' )
ps = acoular.PowerSpectra( time_data=ts, block_size=128, window='Hanning' )
rg = acoular.RectGrid(x_min=-0.6, x_max=-0.0, y_min=-0.3, y_max=0.3, z=0.68,
             increment=0.02)
st = acoular.SteeringVector( grid = rg, mics=mg )
f = acoular.PowerSpectra(time_data=ts,block_size=block)
bf  = acoular.BeamformerBase(freq_data = f,steer= st)


#Integrate function can deal with multiple methods for integration:

#1. a circle containing of three values: x-center, y-center and radius
circle = array([-0.3,-0.1, 0.05])

#2. a rektange containing of 4 values: lower corner(x1, y1) and upper corner(x2, y2).
rect  =  array([-0.5,   -0.15, -0.4 , 0.15])

#3. a polygon containing of vector tuples: x1,y1,x2,y2,...,xi,yi
poly = array([ -0.25, -0.1, -0.1, -0.1, -0.1, -0.2, -0.2, -0.25, -0.3, -0.2])

#4 alternative: define those sectors as Classes
circle_sector = acoular.CircSector(x=-0.3,y= -0.1, r= 0.05)

rect_sector = acoular.RectSector(x_min=-0.5,x_max=-0.4, y_min=-0.15, y_max= 0.15)

#list of points containing x1,y1,x2,y2,...,xi,yi
poly_sector =  acoular.PolySector(edges=[ -0.25, -0.1, -0.1, -0.1, -0.1, -0.2, -0.2, -0.25, -0.3, -0.2])

#multisector allows to sum over multiple different sectors
multi_sector = acoular.MultiSector( sectors = [circle_sector,rect_sector,poly_sector])

#calculate the discrete frequencies for the integration
fftfreqs = arange(block/2+1)*(51200/block)

#integrate SPL values from beamforming results using the shapes
levels_circ = acoular.integrate(bf.result, rg, circle)
levels_rect = acoular.integrate(bf.result, rg, rect)
levels_poly = acoular.integrate(bf.result, rg, poly)

#integrate SPL values from beamforming results using sector classes
levels_circ_sector = acoular.integrate(bf.result, rg, circle_sector)
levels_rekt_sector = acoular.integrate(bf.result, rg, rect_sector)
levels_poly_sector = acoular.integrate(bf.result, rg, poly_sector)
levels_multi_sector = acoular.integrate(bf.result, rg, multi_sector)

#plot map and sectors
figure()
map = bf.synthetic(2000,1)
mx = acoular.L_p(map.max())
imshow(acoular.L_p(map.T), origin='lower', vmin=mx-15,interpolation='nearest', extent=rg.extend(),cmap=cm.hot_r)
colorbar()
circle1 = plt.Circle((-0.3,0.1), 0.05, color='k', fill=False)
plt.gcf().gca().add_artist(circle1)
polygon = Polygon(poly.reshape(-1,2), color='k', fill=False)
plt.gcf().gca().add_artist(polygon)
rect = Rectangle((-0.5,-0.15),0.1,0.3,linewidth=1,edgecolor='k',facecolor='none')
plt.gcf().gca().add_artist(rect)


#plot from shapes
figure()
plot(fftfreqs,acoular.L_p(levels_circ))
plot(fftfreqs,acoular.L_p(levels_rect))
plot(fftfreqs,acoular.L_p(levels_poly))
xlim([2000,20000])
ylim([10,80])
legend(['Circle','Rectangle','Polygon'])

#plot from sector classes
figure()
plot(fftfreqs,acoular.L_p(levels_circ_sector))
plot(fftfreqs,acoular.L_p(levels_rekt_sector))
plot(fftfreqs,acoular.L_p(levels_poly_sector))
plot(fftfreqs,acoular.L_p(levels_multi_sector))
xlim([2000,20000])
ylim([10,80])
legend(['Circle Sector','Rectangle Sector','Polygon Sector','Multisector'])

show()


# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) 2007-2021, Acoular Development Team.
#------------------------------------------------------------------------------
"""
This example shows the different behaviour of SampleSplitter class  
when the maximum size of a block buffer is reached for one object obtaining 
data.

Three different settings can be made by the user:
    * none: no warning, no error
    * warning: a warning appears
    * error: an error is raised
"""

from acoular import TimePower, MaskedTimeSamples, SampleSplitter
import threading
from time import sleep

samples = 25000

# =============================================================================
#  set up data source
# =============================================================================
h5savefile = 'example_data.h5'
ts = MaskedTimeSamples(name=h5savefile,
                       start = 0,
                       stop = samples)  

# =============================================================================
# connect SampleSplitter to data source
# =============================================================================

# set up Sample Splitter
ss = SampleSplitter(source = ts)


# =============================================================================
# create two objects to process the data
# =============================================================================

tp1 = TimePower(source=ss)
tp2 = TimePower(source=ss)

# register these objects at SampleSplitter
ss.register_object(tp1,tp2) # register objects

# =============================================================================
# define functions 
# =============================================================================

def print_number_of_blocks_in_block_buffers():
    """ 
    prints the number of data blocks in SampleSplitter-buffers. For each
    subsequent object, a buffer exist.
    """
    buffers = list(ss.block_buffer.values())
    elements = [len(buf) for buf in buffers]
    print(f"num blocks in buffers: {dict(zip(['tp1','tp2','tp3'], elements))}")


def get_data_fast(obj): # not time consuming function
    """ gets data fast (pause 0.1 seconds)"""
    for _ in obj.result(2048): # 
        print("tp1 calls sample splitter")
        print_number_of_blocks_in_block_buffers()
        sleep(0.1) 

def get_data_slow(obj): # more time consuming function
    """ gets data slow (pause 0.8 seconds)"""
    for i in obj.result(2048): # 
        print("tp3 calls sample splitter")
        print_number_of_blocks_in_block_buffers()
        sleep(0.8) 


# =============================================================================
# prepare and start processing in threads
# (no warning or error when block buffer is full)
# =============================================================================


print("buffer overflow behaviour == 'none'")
print("buffer size is set to a maximum of 5 elements")
ss.buffer_size=5

ss.buffer_overflow_treatment[tp1] = 'none'
ss.buffer_overflow_treatment[tp2] = 'none'

worker1 = threading.Thread(target=get_data_fast, args=(tp1,))
worker2 = threading.Thread(target=get_data_slow, args=(tp2,))

print("start threads")

worker1.start()
worker2.start()

worker1.join()
worker2.join()

print("threads finished")


# =============================================================================
# prepare and start processing in threads
# (only warning when block buffer is full)
# =============================================================================

print("buffer overflow behaviour == 'warning'")
print("buffer size is set to a maximum of 5 elements")
ss.buffer_size=5

ss.buffer_overflow_treatment[tp1] = 'warning'
ss.buffer_overflow_treatment[tp2] = 'warning'

worker1 = threading.Thread(target=get_data_fast, args=(tp1,))
worker2 = threading.Thread(target=get_data_slow, args=(tp2,))

print("start threads")

worker1.start()
worker2.start()

worker1.join()
worker2.join()

print("threads finished")

# =============================================================================
# prepare and start processing in threads
# (raise error when block buffer is full)
# =============================================================================

print("buffer overflow behaviour == 'error'")
print("buffer size is set to a maximum of 5 elements")
ss.buffer_size=5

ss.buffer_overflow_treatment[tp1] = 'error'
ss.buffer_overflow_treatment[tp2] = 'error'

worker1 = threading.Thread(target=get_data_fast, args=(tp1,))
worker2 = threading.Thread(target=get_data_slow, args=(tp2,))

print("start threads")

worker1.start()
worker2.start()

worker1.join()
worker2.join()

print("threads finished")


# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) 2007-2021, Acoular Development Team.
#------------------------------------------------------------------------------
"""
Implements tools for Acoular.

.. autosummary::
    :toctree: generated/
    
    return_result
    barspectrum
    bardata
"""

from traits.api import HasStrictTraits
from numpy import array, concatenate, newaxis, where
from numpy.ma import masked_where
from .spectra import synthetic


def return_result(source, nmax=-1, num=128):
    """
    Collects the output from a 
    :meth:`SamplesGenerator.result()<acoular.tprocess.SamplesGenerator.result>`
    generator and returns an assembled array with all the data.
   
    Parameters
    ----------
    source: SamplesGenerator or derived object.
        This is the  :class:`SamplesGenerator<acoular.tprocess.SamplesGenerator>` data source.
    nmax: integer
        With this parameter, a maximum number of output samples can be set 
        (first dimension of array). If set to -1 (default), samples are 
        collected as long as the generator yields them.
    num : integer
        This parameter defines the size of the blocks that are fetched.
        Defaults to 128.
          
    Returns
    -------
    array of floats (number of samples, source.numchannels)
        Array that holds all the data.
    """
    if nmax > 0: 
        nblocks = (nmax-1) // num + 1
        return concatenate( 
                      list( res for _, res in 
                            zip(range(nblocks), 
                                source.result(num)) ) )[:nmax]
    else:
        return concatenate(list(source.result(num)))



def barspectrum(data, fftfreqs, num = 3, bar = True, xoffset = 0.0):
    """
    Returns synthesized frequency band values of spectral data to be plotted
    as bar graph with the matlpotlib plot command.
    
   
    Parameters
    ----------
    data : array of floats
        The spectral data (sound pressures in Pa) in an array with one value 
        per frequency line.
    fftfreqs : array of floats
        Discrete frequencies from FFT.
    num : integer
        Controls the width of the frequency bands considered; defaults to
        3 (third-octave band).
    bar : bool
        If True, returns bar-like curve. If False, normal plot (direct 
        line between data points) is returned.
    xoffset : float
        If bar is True, offset of the perpendicular line (helpful if 
        plotting several curves above each other). 
        
        ===  =====================
        num  frequency band width
        ===  =====================
        1    octave band
        3    third-octave band
        ===  =====================

    Returns
    -------
    (flulist, plist, fc)
    flulist : array of floats
        Lower/upper band frequencies in plottable format.
    plist : array of floats
        Corresponding synthesized frequency band values in plottable format.
    fc : array of floats
        Evaluated band center frequencies.
    """
    
    if num not in [1,3]:
        print('Only octave and third-octave bands supported at this moment.')
        return (0,0,0)
        

    # preferred center freqs after din en iso 266 for third-octave bands
    fcbase = array([31.5,40,50,63,80,100,125,160,200,250])
    # DIN band center frequencies from 31.5 Hz to 25 kHz
    fc = concatenate((fcbase, fcbase*10., fcbase[:]*100.))[::(3//num)]
    
    
    # exponent for band width calculation
    ep = 1. / (2.*num)
    
    # lowest and highest possible center frequencies 
    # for chosen band and sampling frequency
    f_low = fftfreqs[1]*2**ep
    f_high = fftfreqs[-1]*2**-ep
    # get possible index range
    if fc[0] >= f_low:
        i_low = 0
    else:
        i_low = where(fc < f_low)[0][-1]
    
    if fc[-1] <= f_high:
        i_high = fc.shape[0]
    else:
        i_high = where(fc > f_high)[0][0]
        
    # synthesize sound pressure values
    p = array([ synthetic(data, fftfreqs, list(fc[i_low:i_high]), num) ])
 
    if bar:
        # upper and lower band borders
        flu = concatenate(( fc[i_low:i_low+1]*2**-ep,
                            ( fc[i_low:i_high-1]*2**ep + fc[i_low+1:i_high]*2**-ep ) / 2.,
                            fc[i_high-1:i_high]*2**ep ))
        # band borders as coordinates for bar plotting
        flulist = 2**(2*xoffset*ep) * (array([1,1])[:,newaxis]*flu[newaxis,:]).T.reshape(-1)[1:-1]   
        # sound pressures as list for bar plotting
        plist = (array([1,1])[:,newaxis]*p[newaxis,:]).T.reshape(-1)
    else:
        flulist = fc[i_low:i_high]
        plist = p[0,:]
    #print(flulist.shape, plist.shape)
    return (flulist, plist, fc[i_low:i_high])
    


def bardata(data, fc, num=3, bar = True, xoffset = 0.0, masked = -360):
    """
    Returns data to be plotted
    as bar graph with the matlpotlib plot command.
    
   
    Parameters
    ----------
    data : array of floats
        The spectral data 
    fc : array of floats
        Band center frequencies
    bar : bool
        If True, returns bar-like curve. If False, normal plot (direct 
        line between data points) is returned.
    xoffset : float
        If bar is True, offset of the perpendicular line (helpful if 
        plotting several curves above each other). 
        
        ===  =====================
        num  frequency band width
        ===  =====================
        1    octave band
        3    third-octave band
        ===  =====================

    Returns
    -------
    (flulist, plist)
    flulist : array of floats
        Lower/upper band frequencies in plottable format.
    plist : array of floats
        Corresponding values in plottable format.
    """
    ep = 1. / (2.*num)
    
    if bar:
        # upper and lower band borders
        flu = concatenate(( fc[:1]*2**-ep,
                            ( fc[:-1]*2**ep + fc[1:]*2**-ep ) / 2.,
                            fc[-1:]*2**ep ))
        # band borders as coordinates for bar plotting
        flulist = 2**(xoffset*1./num) * (array([1,1])[:,newaxis]*flu[newaxis,:]).T.reshape(-1)[1:-1]   
        # sound pressures as list for bar plotting
        plist = (array([1,1])[:,newaxis] * data[newaxis,:]).T.reshape(-1)
    else:
        flulist = fc
        plist = data
    #print(flulist.shape, plist.shape)
    if masked > -360:
        plist = masked_where(plist <= masked, plist)
    return (flulist, plist)




# -*- coding: utf-8 -*-
#pylint: disable-msg=E0611, E1101, C0103, R0901, R0902, R0903, R0904, W0232
#------------------------------------------------------------------------------
# Copyright (c) 2007-2021, Acoular Development Team.
#------------------------------------------------------------------------------
"""
This Examples shows how to use the SampleSplitter class in a multithreading 
(3 threads) scenario.

Three different objects (tp1, tp2 and tp3) obtain and process time data 
from the same MaskedTimeSamples object in parallel. The SampleSplitter class
is used to split the data stream for parallel processing.
"""

from acoular import TimePower,MaskedTimeSamples, SampleSplitter
import threading
from time import sleep

samples = 25000

# =============================================================================
#  set up data source
# =============================================================================
h5savefile = 'example_data.h5'
ts = MaskedTimeSamples(name=h5savefile,
                       start = 0,
                       stop = samples)  

# =============================================================================
# connect SampleSplitter to data source
# =============================================================================

# set up Sample Splitter
ss = SampleSplitter(source = ts)


# =============================================================================
# create three objects to process the data
# =============================================================================

tp1 = TimePower(source=ss)
tp2 = TimePower(source=ss)
tp3 = TimePower(source=ss)

# register these objects at SampleSplitter
ss.register_object(tp1,tp2,tp3) # register objects

# =============================================================================
# define functions 
# =============================================================================

def print_number_of_blocks_in_block_buffers():
    """ 
    prints the number of data blocks in SampleSplitter-buffers. For each
    subsequent object, a buffer exist.
    """
    buffers = list(ss.block_buffer.values())
    elements = [len(buf) for buf in buffers]
    print(f"num blocks in buffers: {dict(zip(['tp1','tp2','tp3'], elements))}")

def get_data_fast(obj): # not time consuming function
    """ gets data fast (pause 0.3 seconds)"""
    for _ in obj.result(2048): # 
        print("tp1 calls sample splitter")
        print_number_of_blocks_in_block_buffers()
        sleep(0.3) 

def get_data_mid(obj): # not time consuming function
    """ gets data mid speed (pause 0.5 seconds)"""
    for i in obj.result(2048): # 
        print("tp2 calls sample splitter")
        print_number_of_blocks_in_block_buffers()
        sleep(0.5) 

def get_data_slow(obj): # more time consuming function
    """ gets data slow (pause 0.7 seconds)"""
    for i in obj.result(2048): # 
        print("tp3 calls sample splitter")
        print_number_of_blocks_in_block_buffers()
        sleep(0.7) 

# =============================================================================
# prepare and start processing in threads
# =============================================================================

worker1 = threading.Thread(target=get_data_fast, args=(tp1,))
worker2 = threading.Thread(target=get_data_mid, args=(tp2,))
worker3 = threading.Thread(target=get_data_slow, args=(tp3,))

print("start threads")

worker1.start()
worker2.start()
worker3.start()

[thr.join() for thr in [worker1,worker2,worker3]]

print("threads finished")



# -*- coding: utf-8 -*-
"""
Example "Airfoil in open jet -- Time Beamforming".
acoular example_airfoil_in_open_jet_beamforming.py without frequency beamformers
and only time beamformers

Demonstrates different features of Acoular.

Uses measured data in file example_data.h5,
calibration in file example_calib.xml,
microphone geometry in array_56.xml (part of Acoular).

Copyright (c) 2006-2021 Acoular Development Team.
All rights reserved.
"""

from os import path
import acoular
from pylab import figure, imshow, colorbar, show, subplot,title, tight_layout
from numpy import zeros


micgeofile = path.join(path.split(acoular.__file__)[0],'xml','array_56.xml')
datafile = 'example_data.h5'
calibfile = 'example_calib.xml'
cfreq = 4000

mg = acoular.MicGeom( from_file=micgeofile )
ts = acoular.MaskedTimeSamples( name=datafile )

ts.start = 0
ts.stop = 16000
invalid = [1,7]
ts.invalid_channels = invalid
mg.invalid_channels = invalid
rg = acoular.RectGrid( x_min=-0.6, x_max=0.0, y_min=-0.3, y_max=0.3, z=0.68, \
increment=0.05 )
env = acoular.Environment(c = 346.04)
st = acoular.SteeringVector(grid=rg, mics=mg, env=env)

#===============================================================================
# delay and sum beamformer in time domain
# processing chain: beamforming, filtering, power, average
#===============================================================================
bt = acoular.BeamformerTime(source=ts, steer=st)
ft = acoular.FiltFiltOctave(source=bt, band=cfreq)
pt = acoular.TimePower(source=ft)
avgt = acoular.TimeAverage(source=pt, naverage = 1024)
cacht = acoular.TimeCache( source = avgt) # cache to prevent recalculation

#===============================================================================
# delay and sum beamformer in time domain with autocorrelation removal
# processing chain: zero-phase filtering, beamforming+power, average
#===============================================================================
fi = acoular.FiltFiltOctave(source=ts, band=cfreq)
bts = acoular.BeamformerTimeSq(source=fi, steer=st, r_diag=True)
avgts = acoular.TimeAverage(source=bts, naverage = 1024)
cachts = acoular.TimeCache( source = avgts) # cache to prevent recalculation

#===============================================================================
# clean deconvolution in time domain
# processing chain: zero-phase filtering, clean in time domain, power, average
#===============================================================================
fct = acoular.FiltFiltOctave(source=ts, band=cfreq)
bct = acoular.BeamformerCleant(source=fct, steer=st, n_iter=20,damp=.7)
ptct = acoular.TimePower(source=bct)
avgct = acoular.TimeAverage(source=ptct, naverage = 1024)
cachct = acoular.TimeCache( source = avgct) # cache to prevent recalculation

#===============================================================================
# clean deconvolution in time domain
# processing chain: zero-phase filtering, clean in time domain with 
# autocorrelation removal, average
#===============================================================================
fcts = acoular.FiltFiltOctave(source=ts, band=cfreq)
bcts = acoular.BeamformerCleantSq(source=fcts, steer=st, n_iter=20,damp=.7,r_diag=True)
avgcts = acoular.TimeAverage(source=bcts, naverage = 1024)
cachcts = acoular.TimeCache( source = avgcts) # cache to prevent recalculation

#===============================================================================
# plot result maps for different beamformers in time domain
#===============================================================================
i2 = 1 # no of figure
i1 = 1 #no of subplot
for b in (cacht, cachts, cachct, cachcts):
    # first, plot time-dependent result (block-wise)
    figure(i2,(7,7))
    i2 += 1
    res = zeros(rg.size) # init accumulator for average
    i3 = 1 # no of subplot
    for r in b.result(1):  #one single block
        subplot(4,4,i3)
        i3 += 1
        res += r[0] # average accum.
        map = r[0].reshape(rg.shape)
        mx = acoular.L_p(map.max())
        imshow(acoular.L_p(map.T), vmax=mx, vmin=mx-15, origin='lower',
               interpolation='nearest', extent=rg.extend())
        title('%i' % ((i3-1)*1024))
    res /= i3-1 # average
    tight_layout()

    # second, plot overall result (average over all blocks)
    figure(10)
    subplot(4,4,i1)
    i1 += 1
    map = res.reshape(rg.shape)
    mx = acoular.L_p(map.max())
    imshow(acoular.L_p(map.T), vmax=mx, vmin=mx-15, origin='lower',
           interpolation='nearest', extent=rg.extend())
    colorbar()
    title(('BeamformerTime','BeamformerTimeSq','BeamformerCleant',
           'BeamformerCleantSq')[i2-2])
tight_layout()
# only display result on screen if this script is run directly
if __name__ == '__main__': show()



# -*- coding: utf-8 -*-
"""

"""
from pylab import *
from acoular import *




# files
datafile = 'example_data.h5'

t1 = MaskedTimeSamples(name=datafile)
t1.start = 0 # first sample, default
t1.stop = 16000 # last valid sample = 15999
invalid = [1,7] # list of invalid channels (unwanted microphones etc.)
t1.invalid_channels = invalid 

t2 = ChannelMixer(source=t1)

sig = GenericSignalGenerator(source=t2)

plot(sig.signal())
show()